@@ -188,9 +188,9 @@ from keystoneclient.common import cms
 from keystoneclient import discover
 from keystoneclient import exceptions
 from keystoneclient import session
-from oslo.config import cfg
-from oslo.serialization import jsonutils
-from oslo.utils import timeutils
+from oslo_config import cfg
+from oslo_serialization import jsonutils
+from oslo_utils import timeutils
 import six
 from six.moves import urllib
 
@@ -376,7 +376,7 @@ _DEPRECATED_HEADER_TEMPLATE = {
 }
 
 
-class _BIND_MODE:
+class _BIND_MODE(object):
     DISABLED = 'disabled'
     PERMISSIVE = 'permissive'
     STRICT = 'strict'
@@ -396,22 +396,19 @@ def _get_token_expiration(data):
     if not data:
         raise InvalidToken(_('Token authorization failed'))
     if _token_is_v2(data):
-        timestamp = data['access']['token']['expires']
+        return data['access']['token']['expires']
     elif _token_is_v3(data):
-        timestamp = data['token']['expires_at']
+        return data['token']['expires_at']
     else:
         raise InvalidToken(_('Token authorization failed'))
-    expires = timeutils.parse_isotime(timestamp)
+
+
+def _confirm_token_not_expired(expires):
+    expires = timeutils.parse_isotime(expires)
     expires = timeutils.normalize_time(expires)
-    return expires
-
-
-def _confirm_token_not_expired(data):
-    expires = _get_token_expiration(data)
     utcnow = timeutils.utcnow()
     if utcnow >= expires:
         raise InvalidToken(_('Token authorization failed'))
-    return timeutils.isotime(at=expires, subsecond=True)
 
 
 def _v3_to_v2_catalog(catalog):
@@ -498,6 +495,10 @@ class ConfigurationError(Exception):
     pass
 
 
+class RevocationListError(Exception):
+    pass
+
+
 class _MiniResp(object):
     def __init__(self, error_message, env, headers=[]):
         # The HEAD method is unique: it must never return a body, even if
@@ -572,8 +573,8 @@ class _AuthTokenPlugin(auth.BaseAuthPlugin):
         implementation should use best effort with the information available to
         determine the endpoint.
 
-        :param Session session: The session object that the auth_plugin
-                                belongs to.
+        :param session: The session object that the auth_plugin belongs to.
+        :type session: keystoneclient.session.Session
         :param tuple version: The version number required for this endpoint.
         :param str interface: what visibility the endpoint should have.
 
@@ -666,6 +667,99 @@ class _AuthTokenPlugin(auth.BaseAuthPlugin):
 _AuthTokenPlugin.register_conf_options(CONF, _AUTHTOKEN_GROUP)
 
 
+class _TokenData(object):
+    """An abstraction to show auth_token consumers some of the token contents.
+
+    This is a simplified and cleaned up keystoneclient.access.AccessInfo object
+    with which services relying on auth_token middleware can find details of
+    the current token.
+    """
+
+    def __init__(self, auth_ref):
+        self._stored_auth_ref = auth_ref
+
+    @property
+    def _is_v2(self):
+        return self._stored_auth_ref.version == 'v2.0'
+
+    @property
+    def auth_token(self):
+        """The token data used to authenticate requests.
+
+        :returns: token data.
+        :rtype: str
+        """
+        return self._stored_auth_ref.auth_token
+
+    @property
+    def user_id(self):
+        """The user id associated with the authentication request.
+
+        :rtype: str
+        """
+        return self._stored_auth_ref.user_id
+
+    @property
+    def user_domain_id(self):
+        """Returns the domain id of the user associated with the authentication
+        request.
+
+        :returns: str
+        """
+        # NOTE(jamielennox): v2 AccessInfo returns 'default' for domain_id
+        # because it can't know that value. We want to return None instead.
+        if self._is_v2:
+            return None
+
+        return self._stored_auth_ref.user_domain_id
+
+    @property
+    def project_id(self):
+        """The project ID associated with the authentication.
+
+        :rtype: str
+        """
+        return self._stored_auth_ref.project_id
+
+    @property
+    def project_domain_id(self):
+        """The domain id of the project associated with the authentication
+        request.
+
+        :rtype: str
+        """
+        # NOTE(jamielennox): v2 AccessInfo returns 'default' for domain_id
+        # because it can't know that value. We want to return None instead.
+        if self._is_v2:
+            return None
+
+        return self._stored_auth_ref.project_domain_id
+
+    @property
+    def trust_id(self):
+        """Returns the trust id associated with the authentication request..
+
+        :rtype: str
+        """
+        return self._stored_auth_ref.trust_id
+
+    @property
+    def role_ids(self):
+        """Role ids of the user associated with the authentication request.
+
+        :rtype: set(str)
+        """
+        return frozenset(self._stored_auth_ref.role_ids or [])
+
+    @property
+    def role_names(self):
+        """Role names of the user associated with the authentication request.
+
+        :rtype: set(str)
+        """
+        return frozenset(self._stored_auth_ref.role_names or [])
+
+
 class _UserAuthPlugin(base_identity.BaseIdentityPlugin):
     """The incoming authentication credentials.
 
@@ -677,15 +771,56 @@ class _UserAuthPlugin(base_identity.BaseIdentityPlugin):
     authentication plugin when communicating via a session.
     """
 
-    def __init__(self, auth_ref):
+    def __init__(self, user_auth_ref, serv_auth_ref):
         super(_UserAuthPlugin, self).__init__(reauthenticate=False)
-        self._stored_auth_ref = auth_ref
+        self._user_auth_ref = user_auth_ref
+        self._serv_auth_ref = serv_auth_ref
+        self._user_data = None
+        self._serv_data = None
+
+    @property
+    def has_user_token(self):
+        """Did this authentication request contained a user auth token."""
+        return self._user_auth_ref is not None
+
+    @property
+    def user(self):
+        """Authentication information about the user token.
+
+        Will return None if a user token was not passed with this request.
+        """
+        if not self.has_user_token:
+            return None
+
+        if not self._user_data:
+            self._user_data = _TokenData(self._user_auth_ref)
+
+        return self._user_data
+
+    @property
+    def has_service_token(self):
+        """Did this authentication request contained a service token."""
+        return self._serv_auth_ref is not None
+
+    @property
+    def service(self):
+        """Authentication information about the service token.
+
+        Will return None if a user token was not passed with this request.
+        """
+        if not self.has_service_token:
+            return None
+
+        if not self._serv_data:
+            self._serv_data = _TokenData(self._serv_auth_ref)
+
+        return self._serv_data
 
     def get_auth_ref(self, session, **kwargs):
         # NOTE(jamielennox): We will always use the auth_ref that was
         # calculated by the middleware. reauthenticate=False in __init__ should
         # ensure that this function is only called on the first access.
-        return self._stored_auth_ref
+        return self._user_auth_ref
 
 
 class AuthProtocol(object):
@@ -702,14 +837,23 @@ class AuthProtocol(object):
 
         # delay_auth_decision means we still allow unauthenticated requests
         # through and we let the downstream service make the final decision
-        self._delay_auth_decision = (self._conf_get('delay_auth_decision') in
-                                     (True, 'true', 't', '1', 'on', 'yes', 'y')
-                                     )
-
+        self._delay_auth_decision = self._conf_get('delay_auth_decision')
         self._include_service_catalog = self._conf_get(
             'include_service_catalog')
 
-        self._identity_server_obj = None
+        self._identity_server = self._create_identity_server()
+
+        self._auth_uri = self._conf_get('auth_uri')
+        if not self._auth_uri:
+            self._LOG.warning(
+                _LW('Configuring auth_uri to point to the public identity '
+                    'endpoint is required; clients may not be able to '
+                    'authenticate against an admin endpoint'))
+
+            # FIXME(dolph): drop support for this fallback behavior as
+            # documented in bug 1207517.
+
+            self._auth_uri = self._identity_server.auth_uri
 
         # signing
         self._signing_dirname = self._conf_get('signing_dir')
@@ -767,26 +911,30 @@ class AuthProtocol(object):
         def _fmt_msg(env):
             msg = ('user: user_id %s, project_id %s, roles %s '
                    'service: user_id %s, project_id %s, roles %s' % (
-                   env.get('X_USER_ID'), env.get('X_PROJECT_ID'),
-                   env.get('X_ROLES'), env.get('X_SERVICE_USER_ID'),
-                   env.get('X_SERVICE_PROJECT_ID'),
-                   env.get('X_SERVICE_ROLES')))
+                       env.get('HTTP_X_USER_ID'), env.get('HTTP_X_PROJECT_ID'),
+                       env.get('HTTP_X_ROLES'),
+                       env.get('HTTP_X_SERVICE_USER_ID'),
+                       env.get('HTTP_X_SERVICE_PROJECT_ID'),
+                       env.get('HTTP_X_SERVICE_ROLES')))
             return msg
 
         self._token_cache.initialize(env)
         self._remove_auth_headers(env)
 
         try:
+            user_auth_ref = None
+            serv_auth_ref = None
 
             try:
                 self._LOG.debug('Authenticating user token')
                 user_token = self._get_user_token_from_header(env)
-                token_info = self._validate_token(user_token, env)
-                auth_ref = access.AccessInfo.factory(body=token_info,
-                                                     auth_token=user_token)
-                env['keystone.token_info'] = token_info
-                env['keystone.token_auth'] = _UserAuthPlugin(auth_ref)
-                user_headers = self._build_user_headers(auth_ref, token_info)
+                user_token_info = self._validate_token(user_token, env)
+                user_auth_ref = access.AccessInfo.factory(
+                    body=user_token_info,
+                    auth_token=user_token)
+                env['keystone.token_info'] = user_token_info
+                user_headers = self._build_user_headers(user_auth_ref,
+                                                        user_token_info)
                 self._add_headers(env, user_headers)
             except InvalidToken:
                 if self._delay_auth_decision:
@@ -805,6 +953,9 @@ class AuthProtocol(object):
                 if serv_token is not None:
                     serv_token_info = self._validate_token(
                         serv_token, env)
+                    serv_auth_ref = access.AccessInfo.factory(
+                        body=serv_token_info,
+                        auth_token=serv_token)
                     serv_headers = self._build_service_headers(serv_token_info)
                     self._add_headers(env, serv_headers)
             except InvalidToken:
@@ -814,16 +965,14 @@ class AuthProtocol(object):
                     _LI('Invalid service token - rejecting request'))
                 return self._reject_request(env, start_response)
 
-        except exceptions.NoMatchingPlugin as e:
-            msg = _LC('Required auth plugin does not exist. %s') % e
-            self._LOG.critical(msg)
-            return self._do_503_error(env, start_response)
+            env['keystone.token_auth'] = _UserAuthPlugin(user_auth_ref,
+                                                         serv_auth_ref)
 
         except ServiceError as e:
             self._LOG.critical(_LC('Unable to obtain admin token: %s'), e)
             return self._do_503_error(env, start_response)
 
-        self._LOG.debug("Received request from %s" % _fmt_msg(env))
+        self._LOG.debug("Received request from %s", _fmt_msg(env))
 
         return self._call_app(env, start_response)
 
@@ -869,8 +1018,8 @@ class AuthProtocol(object):
         """Get token id from request.
 
         :param env: wsgi request environment
-        :return token id
-        :raises InvalidToken if no token is provided in request
+        :returns: token id
+        :raises InvalidToken: if no token is provided in request
 
         """
         token = self._get_header(env, 'X-Auth-Token',
@@ -888,14 +1037,14 @@ class AuthProtocol(object):
         """Get service token id from request.
 
         :param env: wsgi request environment
-        :return service token id or None if not present
+        :returns: service token id or None if not present
 
         """
         return self._get_header(env, 'X-Service-Token')
 
     @property
     def _reject_auth_headers(self):
-        header_val = 'Keystone uri=\'%s\'' % self._identity_server.auth_uri
+        header_val = 'Keystone uri=\'%s\'' % self._auth_uri
         return [('WWW-Authenticate', header_val)]
 
     def _reject_request(self, env, start_response):
@@ -903,7 +1052,7 @@ class AuthProtocol(object):
 
         :param env: wsgi request environment
         :param start_response: wsgi response callback
-        :returns HTTPUnauthorized http response
+        :returns: HTTPUnauthorized http response
 
         """
         resp = _MiniResp('Authentication required',
@@ -915,10 +1064,10 @@ class AuthProtocol(object):
         """Authenticate user token
 
         :param token: token id
+        :param env: wsgi environment
         :param retry: Ignored, as it is not longer relevant
-        :return uncrypted body of the token if the token is valid
-        :raise InvalidToken if token is rejected
-        :no longer raises ServiceError since it no longer makes RPC
+        :returns: uncrypted body of the token if the token is valid
+        :raises InvalidToken: if token is rejected
 
         """
         token_id = None
@@ -947,19 +1096,27 @@ class AuthProtocol(object):
                                 _('Token authorization failed'))
                 self._confirm_token_bind(data, env)
             else:
+                verified = None
                 # Token wasn't cached. In this case, the token needs to be
                 # checked that it's not expired, and also put in the cache.
-                if cms.is_pkiz(token):
-                    verified = self._verify_pkiz_token(token, token_ids)
+                try:
+                    if cms.is_pkiz(token):
+                        verified = self._verify_pkiz_token(token, token_ids)
+                    elif cms.is_asn1_token(token):
+                        verified = self._verify_signed_token(token, token_ids)
+                except exceptions.CertificateConfigError:
+                    self._LOG.warn(_LW('Fetch certificate config failed, '
+                                       'fallback to online validation.'))
+                except RevocationListError:
+                    self._LOG.warn(_LW('Fetch revocation list failed, '
+                                       'fallback to online validation.'))
+
+                if verified is not None:
                     data = jsonutils.loads(verified)
-                    expires = _confirm_token_not_expired(data)
-                elif cms.is_asn1_token(token):
-                    verified = self._verify_signed_token(token, token_ids)
-                    data = jsonutils.loads(verified)
-                    expires = _confirm_token_not_expired(data)
+                    expires = _get_token_expiration(data)
+                    _confirm_token_not_expired(expires)
                 else:
-                    data = self._identity_server.verify_token(token,
-                                                              retry)
+                    data = self._identity_server.verify_token(token, retry)
                     # No need to confirm token expiration here since
                     # verify_token fails for expired tokens.
                     expires = _get_token_expiration(data)
@@ -970,6 +1127,8 @@ class AuthProtocol(object):
             self._LOG.debug('Token validation failure.', exc_info=True)
             self._LOG.warn(_LW('Authorization failed for token'))
             raise InvalidToken(_('Token authorization failed'))
+        except ServiceError:
+            raise
         except Exception:
             self._LOG.debug('Token validation failure.', exc_info=True)
             if token_id:
@@ -985,7 +1144,7 @@ class AuthProtocol(object):
 
         :param token_info: token object returned by identity
                            server on authentication
-        :raise InvalidToken: when unable to parse token object
+        :raises InvalidToken: when unable to parse token object
 
         """
         roles = ','.join(auth_ref.role_names)
@@ -1022,7 +1181,7 @@ class AuthProtocol(object):
 
         :param token_info: token object returned by identity
                            server on authentication
-        :raise InvalidToken: when unable to parse token object
+        :raises InvalidToken: when unable to parse token object
 
         """
         auth_ref = access.AccessInfo.factory(body=token_info)
@@ -1045,7 +1204,7 @@ class AuthProtocol(object):
         """Convert header to wsgi env variable.
 
         :param key: http header name (ex. 'X-Auth-Token')
-        :return wsgi env variable name (ex. 'HTTP_X_AUTH_TOKEN')
+        :returns: wsgi env variable name (ex. 'HTTP_X_AUTH_TOKEN')
 
         """
         return 'HTTP_%s' % key.replace('-', '_').upper()
@@ -1223,8 +1382,8 @@ class AuthProtocol(object):
             current_mode = stat.S_IMODE(os.stat(self._signing_dirname).st_mode)
             if current_mode != stat.S_IRWXU:
                 self._LOG.warning(
-                    _LW('signing_dir mode is %s instead of %s'),
-                    oct(current_mode), oct(stat.S_IRWXU))
+                    _LW('signing_dir mode is %(mode)s instead of %(need)s'),
+                    {'mode': oct(current_mode), 'need': oct(stat.S_IRWXU)})
         else:
             os.makedirs(self._signing_dirname, stat.S_IRWXU)
 
@@ -1306,86 +1465,59 @@ class AuthProtocol(object):
             self._signing_ca_file_name,
             self._identity_server.fetch_ca_cert())
 
-    @property
-    def _identity_server(self):
-        if not self._identity_server_obj:
-            # NOTE(jamielennox): Loading Session here should be exactly the
-            # same as calling Session.load_from_conf_options(CONF, GROUP)
-            # however we can't do that because we have to use _conf_get to
-            # support the paste.ini options.
-            sess = session.Session.construct(dict(
-                cert=self._conf_get('certfile'),
-                key=self._conf_get('keyfile'),
-                cacert=self._conf_get('cafile'),
-                insecure=self._conf_get('insecure'),
-                timeout=self._conf_get('http_connect_timeout')
-            ))
+    def _create_identity_server(self):
+        # NOTE(jamielennox): Loading Session here should be exactly the
+        # same as calling Session.load_from_conf_options(CONF, GROUP)
+        # however we can't do that because we have to use _conf_get to
+        # support the paste.ini options.
+        sess = session.Session.construct(dict(
+            cert=self._conf_get('certfile'),
+            key=self._conf_get('keyfile'),
+            cacert=self._conf_get('cafile'),
+            insecure=self._conf_get('insecure'),
+            timeout=self._conf_get('http_connect_timeout')
+        ))
 
-            # NOTE(jamielennox): The original auth mechanism allowed deployers
-            # to configure authentication information via paste file. These
-            # are accessible via _conf_get, however this doesn't work with the
-            # plugin loading mechanisms. For using auth plugins we only support
-            # configuring via the CONF file.
-            auth_plugin = auth.load_from_conf_options(CONF, _AUTHTOKEN_GROUP)
+        # NOTE(jamielennox): The original auth mechanism allowed deployers
+        # to configure authentication information via paste file. These
+        # are accessible via _conf_get, however this doesn't work with the
+        # plugin loading mechanisms. For using auth plugins we only support
+        # configuring via the CONF file.
+        auth_plugin = auth.load_from_conf_options(CONF, _AUTHTOKEN_GROUP)
 
-            if not auth_plugin:
-                # NOTE(jamielennox): Loading AuthTokenPlugin here should be
-                # exactly the same as calling
-                # _AuthTokenPlugin.load_from_conf_options(CONF, GROUP) however
-                # we can't do that because we have to use _conf_get to support
-                # the paste.ini options.
-                auth_plugin = _AuthTokenPlugin.load_from_options(
-                    auth_host=self._conf_get('auth_host'),
-                    auth_port=int(self._conf_get('auth_port')),
-                    auth_protocol=self._conf_get('auth_protocol'),
-                    auth_admin_prefix=self._conf_get('auth_admin_prefix'),
-                    admin_user=self._conf_get('admin_user'),
-                    admin_password=self._conf_get('admin_password'),
-                    admin_tenant_name=self._conf_get('admin_tenant_name'),
-                    admin_token=self._conf_get('admin_token'),
-                    identity_uri=self._conf_get('identity_uri'),
-                    log=self._LOG)
+        if not auth_plugin:
+            # NOTE(jamielennox): Loading AuthTokenPlugin here should be
+            # exactly the same as calling
+            # _AuthTokenPlugin.load_from_conf_options(CONF, GROUP) however
+            # we can't do that because we have to use _conf_get to support
+            # the paste.ini options.
+            auth_plugin = _AuthTokenPlugin.load_from_options(
+                auth_host=self._conf_get('auth_host'),
+                auth_port=int(self._conf_get('auth_port')),
+                auth_protocol=self._conf_get('auth_protocol'),
+                auth_admin_prefix=self._conf_get('auth_admin_prefix'),
+                admin_user=self._conf_get('admin_user'),
+                admin_password=self._conf_get('admin_password'),
+                admin_tenant_name=self._conf_get('admin_tenant_name'),
+                admin_token=self._conf_get('admin_token'),
+                identity_uri=self._conf_get('identity_uri'),
+                log=self._LOG)
 
-            adap = adapter.Adapter(
-                sess,
-                auth=auth_plugin,
-                service_type='identity',
-                interface='admin',
-                connect_retries=self._conf_get('http_request_max_retries'))
+        adap = adapter.Adapter(
+            sess,
+            auth=auth_plugin,
+            service_type='identity',
+            interface='admin',
+            connect_retries=self._conf_get('http_request_max_retries'))
 
-            server_class = self._get_server_class(adap)
-            adap.version = server_class.AUTH_VERSION
-
-            self._identity_server_obj = server_class(
-                self._LOG,
-                adap,
-                include_service_catalog=self._include_service_catalog,
-                auth_uri=self._conf_get('auth_uri'))
-
-        return self._identity_server_obj
-
-    def _get_server_class(self, adap):
-        # FIXME(jamielennox): Checking string equality is bad, but consistent
-        # with the existing code. Fix this to better handle selecting v3.
         auth_version = self._conf_get('auth_version')
-
-        if auth_version == 'v3.0':
-            return _V3IdentityServer
-        elif auth_version:
-            return _V2IdentityServer
-
-        for klass in _VERSIONS_TO_ATTEMPT:
-            if adap.get_endpoint(version=klass.AUTH_VERSION):
-                msg = _LI('Auth Token confirmed use of %s apis')
-                self._LOG.info(msg, auth_version)
-                return klass
-
-        versions = ['v%d.%d' % s.AUTH_VERSION for s in _VERSIONS_TO_ATTEMPT]
-        self._LOG.error(_LE('No attempted versions [%s] supported by server') %
-                        ', '.join(versions))
-
-        msg = _('No compatible apis supported by server')
-        raise ServiceError(msg)
+        if auth_version is not None:
+            auth_version = discover.normalize_version_number(auth_version)
+        return _IdentityServer(
+            self._LOG,
+            adap,
+            include_service_catalog=self._include_service_catalog,
+            requested_auth_version=auth_version)
 
     def _token_cache_factory(self):
         security_strategy = self._conf_get('memcache_security_strategy')
@@ -1481,30 +1613,69 @@ class _IdentityServer(object):
 
     """
 
-    AUTH_VERSION = None
-
-    def __init__(self, log, adap, include_service_catalog=None, auth_uri=None):
+    def __init__(self, log, adap, include_service_catalog=None,
+                 requested_auth_version=None):
         self._LOG = log
         self._adapter = adap
         self._include_service_catalog = include_service_catalog
+        self._requested_auth_version = requested_auth_version
 
-        if auth_uri is None:
-            self._LOG.warning(
-                _LW('Configuring auth_uri to point to the public identity '
-                    'endpoint is required; clients may not be able to '
-                    'authenticate against an admin endpoint'))
+        # Built on-demand with self._request_strategy.
+        self._request_strategy_obj = None
 
-            # FIXME(dolph): drop support for this fallback behavior as
-            # documented in bug 1207517.
-            auth_uri = adap.get_endpoint(interface=auth.AUTH_INTERFACE)
+    @property
+    def auth_uri(self):
+        auth_uri = self._adapter.get_endpoint(interface=auth.AUTH_INTERFACE)
 
-            # NOTE(jamielennox): This weird stripping of the prefix hack is
-            # only relevant to the legacy case. We urljoin '/' to get just the
-            # base URI as this is the original behaviour.
-            if isinstance(adap.auth, _AuthTokenPlugin):
-                auth_uri = urllib.parse.urljoin(auth_uri, '/').rstrip('/')
+        # NOTE(jamielennox): This weird stripping of the prefix hack is
+        # only relevant to the legacy case. We urljoin '/' to get just the
+        # base URI as this is the original behaviour.
+        if isinstance(self._adapter.auth, _AuthTokenPlugin):
+            auth_uri = urllib.parse.urljoin(auth_uri, '/').rstrip('/')
 
-        self.auth_uri = auth_uri
+        return auth_uri
+
+    @property
+    def auth_version(self):
+        return self._request_strategy.AUTH_VERSION
+
+    @property
+    def _request_strategy(self):
+        if not self._request_strategy_obj:
+            strategy_class = self._get_strategy_class()
+            self._adapter.version = strategy_class.AUTH_VERSION
+
+            self._request_strategy_obj = strategy_class(
+                self._json_request,
+                self._adapter,
+                include_service_catalog=self._include_service_catalog)
+
+        return self._request_strategy_obj
+
+    def _get_strategy_class(self):
+        if self._requested_auth_version:
+            # A specific version was requested.
+            if discover.version_match(_V3RequestStrategy.AUTH_VERSION,
+                                      self._requested_auth_version):
+                return _V3RequestStrategy
+
+            # The version isn't v3 so we don't know what to do. Just assume V2.
+            return _V2RequestStrategy
+
+        # Specific version was not requested then we fall through to
+        # discovering available versions from the server
+        for klass in _REQUEST_STRATEGIES:
+            if self._adapter.get_endpoint(version=klass.AUTH_VERSION):
+                msg = _LI('Auth Token confirmed use of %s apis')
+                self._LOG.info(msg, self._requested_auth_version)
+                return klass
+
+        versions = ['v%d.%d' % s.AUTH_VERSION for s in _REQUEST_STRATEGIES]
+        self._LOG.error(_LE('No attempted versions [%s] supported by server'),
+                        ', '.join(versions))
+
+        msg = _('No compatible apis supported by server')
+        raise ServiceError(msg)
 
     def verify_token(self, user_token, retry=True):
         """Authenticate user token with identity server.
@@ -1513,21 +1684,21 @@ class _IdentityServer(object):
         :param retry: flag that forces the middleware to retry
                       user authentication when an indeterminate
                       response is received. Optional.
-        :return: token object received from identity server on success
-        :raise InvalidToken: if token is rejected
-        :raise ServiceError: if unable to authenticate token
+        :returns: token object received from identity server on success
+        :raises InvalidToken: if token is rejected
+        :raises ServiceError: if unable to authenticate token
 
         """
         user_token = _safe_quote(user_token)
 
         try:
-            response, data = self._do_verify_token(user_token)
+            response, data = self._request_strategy.verify_token(user_token)
         except exceptions.NotFound as e:
             self._LOG.warn(_LW('Authorization failed for token'))
-            self._LOG.warn(_LW('Identity response: %s') % e.response.text)
+            self._LOG.warn(_LW('Identity response: %s'), e.response.text)
         except exceptions.Unauthorized as e:
             self._LOG.info(_LI('Identity server rejected authorization'))
-            self._LOG.warn(_LW('Identity response: %s') % e.response.text)
+            self._LOG.warn(_LW('Identity response: %s'), e.response.text)
             if retry:
                 self._LOG.info(_LI('Retrying validation'))
                 return self.verify_token(user_token, False)
@@ -1535,7 +1706,7 @@ class _IdentityServer(object):
             self._LOG.error(
                 _LE('Bad response code while validating token: %s'),
                 e.http_status)
-            self._LOG.warn(_LW('Identity response: %s') % e.response.text)
+            self._LOG.warn(_LW('Identity response: %s'), e.response.text)
         else:
             if response.status_code == 200:
                 return data
@@ -1549,12 +1720,14 @@ class _IdentityServer(object):
                 authenticated=True,
                 endpoint_filter={'version': (2, 0)})
         except exceptions.HTTPError as e:
-            raise ServiceError(_('Failed to fetch token revocation list: %d') %
-                               e.http_status)
+            raise RevocationListError(_('Failed to fetch token revocation '
+                                        'list: %d') % e.http_status)
         if response.status_code != 200:
-            raise ServiceError(_('Unable to fetch token revocation list.'))
+            raise RevocationListError(_('Unable to fetch token revocation '
+                                        'list.'))
         if 'signed' not in data:
-            raise ServiceError(_('Revocation list improperly formatted.'))
+            raise RevocationListError(_('Revocation list improperly '
+                                        'formatted.'))
         return data['signed']
 
     def fetch_signing_cert(self):
@@ -1569,8 +1742,8 @@ class _IdentityServer(object):
         :param method: http method
         :param path: relative request url
         :param **kwargs: additional parameters used by session or endpoint
-        :return (http response object, response body parsed as json)
-        :raise ServerError when unable to communicate with identity server.
+        :returns: http response object, response body parsed as json
+        :raises ServerError: when unable to communicate with identity server.
 
         """
         headers = kwargs.setdefault('headers', {})
@@ -1588,7 +1761,7 @@ class _IdentityServer(object):
 
     def _fetch_cert_file(self, cert_type):
         try:
-            response = self._do_fetch_cert_file(cert_type)
+            response = self._request_strategy.fetch_cert_file(cert_type)
         except exceptions.HTTPError as e:
             raise exceptions.CertificateConfigError(e.details)
         if response.status_code != 200:
@@ -1596,25 +1769,41 @@ class _IdentityServer(object):
         return response.text
 
 
-class _V2IdentityServer(_IdentityServer):
+class _RequestStrategy(object):
+
+    AUTH_VERSION = None
+
+    def __init__(self, json_request, adap, include_service_catalog=None):
+        self._json_request = json_request
+        self._adapter = adap
+        self._include_service_catalog = include_service_catalog
+
+    def verify_token(self, user_token):
+        pass
+
+    def fetch_cert_file(self, cert_type):
+        pass
+
+
+class _V2RequestStrategy(_RequestStrategy):
 
     AUTH_VERSION = (2, 0)
 
-    def _do_verify_token(self, user_token):
+    def verify_token(self, user_token):
         return self._json_request('GET',
                                   '/tokens/%s' % user_token,
                                   authenticated=True)
 
-    def _do_fetch_cert_file(self, cert_type):
+    def fetch_cert_file(self, cert_type):
         return self._adapter.get('/certificates/%s' % cert_type,
                                  authenticated=False)
 
 
-class _V3IdentityServer(_IdentityServer):
+class _V3RequestStrategy(_RequestStrategy):
 
     AUTH_VERSION = (3, 0)
 
-    def _do_verify_token(self, user_token):
+    def verify_token(self, user_token):
         path = '/auth/tokens'
         if not self._include_service_catalog:
             path += '?nocatalog'
@@ -1624,7 +1813,7 @@ class _V3IdentityServer(_IdentityServer):
                                   authenticated=True,
                                   headers={'X-Subject-Token': user_token})
 
-    def _do_fetch_cert_file(self, cert_type):
+    def fetch_cert_file(self, cert_type):
         if cert_type == 'signing':
             cert_type = 'certificates'
 
@@ -1632,6 +1821,10 @@ class _V3IdentityServer(_IdentityServer):
                                  authenticated=False)
 
 
+# NOTE(jamielennox): must be defined after request strategy classes
+_REQUEST_STRATEGIES = [_V3RequestStrategy, _V2RequestStrategy]
+
+
 class _TokenCache(object):
     """Encapsulates the auth_token token cache functionality.
 
@@ -1937,10 +2130,6 @@ def app_factory(global_conf, **local_conf):
     return AuthProtocol(None, conf)
 
 
-# NOTE(jamielennox): must be defined after identity server classes
-_VERSIONS_TO_ATTEMPT = [_V3IdentityServer, _V2IdentityServer]
-
-
 if __name__ == '__main__':
     def echo_app(environ, start_response):
         """A WSGI application that echoes the CGI environment to the user."""
